
//========================= SHT HELPER FUNCTIONS ===============================
//
// Functions to provide CRC check, integer only and floating point conversion
// of sensor output into real units (deg C and % RH).
//
// Grahame Marsh - Highlands Scotland
//
// You get this for free - I make no promises
//
//==============================================================================


//--- CRC Check ----------------------------------------------------------------

// This implements the table method to check the CRC returned from the sensor for
// either the temperature or humidity data.
// The helper function:
// function IsSHTCRCGood (Status : byte; Command : TSHTCmds; Value : word; TestCRC : byte) : boolean;
//
// Status  = the status byte returned from the sensor
// Command = either scTemp or scHumidity depending on which CRC is being checked
// Value   = the data value returned from the sensor
// TestCRC = the CRC value returned from the sensor
//
// The function returns true if the TestCRC and the calculated CRC are equal

// Put the CRC data table into EEPROM ... or you put it where you want it ...
{$EEPROM}

{$IFDEF SHTCRC}  // prevents inclusion - use ($DEFINE SHTCRC} ahead of {$I
structconst
  CRCTable : Table [0..255] of byte = (

  0,  49,  98,  83, 196, 245, 166, 151, 185, 136, 219, 234, 125,  76,  31,  46,
 67, 114,  33,  16, 135, 182, 229, 212, 250, 203, 152, 169,  62,  15,  92, 109,
134, 183, 228, 213,  66, 115,  32,  17,  63,  14,  93, 108, 251, 202, 153, 168,
197, 244, 167, 150,   1,  48,  99,  82, 124,  77,  30,  47, 184, 137, 218, 235,
 61,  12,  95, 110, 249, 200, 155, 170, 132, 181, 230, 215,  64, 113,  34,  19,
126,  79,  28,  45, 186, 139, 216, 233, 199, 246, 165, 148,   3,  50,  97,  80,
187, 138, 217, 232, 127,  78,  29,  44,   2,  51,  96,  81, 198, 247, 164, 149,
248, 201, 154, 171,  60,  13,  94, 111,  65, 112,  35,  18, 133, 180, 231, 214,
122,  75,  24,  41, 190, 143, 220, 237, 195, 242, 161, 144,   7,  54, 101,  84,
 57,   8,  91, 106, 253, 204, 159, 174, 128, 177, 226, 211,  68, 117,  38,  23,
252, 205, 158, 175,  56,   9,  90, 107,  69, 116,  39,  22, 129, 176, 227, 210,
191, 142, 221, 236, 123,  74,  25,  40,   6,  55, 100,  85, 194, 243, 160, 145,
 71, 118,  37,  20, 131, 178, 225, 208, 254, 207, 156, 173,  58,  11,  88, 105,
  4,  53, 102,  87, 192, 241, 162, 147, 189, 140, 223, 238, 121,  72,  27,  42,
193, 240, 163, 146,   5,  52, 103,  86, 120,  73,  26,  43, 188, 141, 222, 239,
130, 179, 224, 209,  70, 119,  36,  21,  59,  10,  89, 104, 255, 206, 157, 172);

{$IDATA}

type
  TSHTCmds = (scTemp = 3, scHumidity = 5);

// internal working variable - do not modify
  var
    __CRC : byte;
  
  procedure AddCRC (X : byte);
  begin
    __CRC := GetTable (CRCTable, __CRC xor X);
  end;

function IsSHTCRCGood (Status : byte; Command : TSHTCmds; Value : word; TestCRC : byte) : boolean;
begin
  __CRC := Mirror8 (Status and $0F);
  AddCRC (ord(Command));
  AddCRC (Hi (Value));
  AddCRC (Lo (Value));
  return (TestCRC = Mirror8 (__CRC));
end;
{$ENDIF}

(* Example:
var
  SOT : word;
...
  repeat
    SHT11StartTemp;
    repeat
    until SHT11ConvState;
    SOT := SHT11GetTemp;
  until IsSHTCRCGood (SHT11GetStatus, scTemp, SOT, SHT11CRC);
... now use the checked value of SOT
*)

//--- INTEGER CONVERSION HELPER FUNCTIONS --------------------------------------
// These helper functions use only integer arithemetic to convert the sensor
// values into meaningful number either deg C or %RH.  Conversion into deg F is
// not supported here but easy to add.  Returned values are 10x the actual value,
// so for example 245 would be, say, 24.5 deg C or 24.5 % RH.

//--- LOW RESOLUTION CONVERSIONS (8 bit humidity, 12 bit temperature) ----------
{$IDATA}

// ==== HUMIDITY ====

// Humidity by simplest linear conversion +/- 2.2% RH accuracy
// implemented by RHsimple = c1 + c2 x SORH
// where c1=0.5; c2=0.5
function HumidityLo1 (Value : word) : word;
begin
  return (5 * Value + 5);
end;

// Humidity by 2*linear conversion +/- 0.8% RH
// implemented by RH = (a x SORH + b ) /256
// Validity            a      b
// 0 <= SORH <=107     143    -512
// 108 <= SO <=255     111    2893
function HumidityLo2 (Value : word) : word;
begin
  if Value <= 107 then
    Value := 143 * Value;
    if Value < 512 then
      return (0);      // 0% RH
    else
      Value := Value - 512;
    endif;
  else
    Value := 111 * Value + 2893;
    if Value > 25600 then
      return (1000);   // 100% RH
    endif;
  endif;
  return (MulDivInt (Value, 10, 256));
end;

// Humidity by 2nd order polynominal +/- 0.1% RH
// implemented by RH = c1 + c2 x SORH + c3 x SORH^2
// where c1=-4; c2=0.648; c3=-7.2*10^-4
function HumidityLo3 (Value : word) : word;
begin
  return (-MulDivInt (Value, Value, 139) + MulDivInt (Value, 648, 100) - 40);
end;

// Humidity - temperature compensated
// implemented by RH = (T-25) x (t1 + t2 * SORH) + RH
// where t1=0.01; t2=0.00128; RH calculated using above formula 3
function HumidityLoComp (Value,  Temp : word) : word;
begin
  return (HumidityLo3 (Value) + MulDivInt(Temp - 25, 1 + MulDivInt (Value, 100, 781), 10));
end;

// ==== Temperature ====

// Linear temperature conversion
// implemented by T = d1 + d2 x SOT
// where d1=-40; d2=0.04
function TemperatureLo (Value : word) : integer;
begin
  return (MulDivInt (Value, 4, 10) - 400);
end;

// Temperature - non linearity compensated
// implemented by T = d1 + d2 x SOT + d3 x (SOT - f)^2
// where d1=-40; d2=0.04; d3=-3.2e-7; f=1750
function TemperatureLoComp (Value : word) : integer;
var
  I : integer;
begin
  I := abs(integer (Value) - 1750);
  return (TemperatureLo (Value) - MulDivInt (I, I, 31250) div 10);
end;

//--- HIGH RESOLUTION (12 bit humidity, 14 bit temperature)

// ==== Humidity ====

// Humidity by 2nd order polynominal +/- 0.1% RH
// implemented by RH = c1 + c2 x SORH + c3 x SORH^2
// where c1=-4; c2=0.0405; c3=-2.8*10^-6
function HumidityHi (Value : word) : word;
begin
  return (-MulDivInt (Value, Value, 35714) + MulDivInt (Value, 405, 1000) - 40);
end;

// Humidity - non linearity compensated
// implemented by RH = (T-25) x (t1 + t2 * SORH) + RH
// where t1=0.01; t2=0.00008; RH calculated using above formula
function HumidityHiComp (Value,  Temp : word) : word;
begin
  return (HumidityHi (Value) + MulDivInt(Temp - 25, 1 + (Value div 125), 10));
end;

// ==== Temperature ====

// Linear temperature conversion
// implemented by T = d1 + d2 x SOT
// where d1=-40; d2=0.01
function TemperatureHi (Value : word) : integer;
begin
  return (integer (Value div 10) - 400);
end;

// Temperature - non linearity compensated
// implemented by T = d1 + d2 x SOT + d3 x (SOT - f)^2
// where d1=-40; d2=0.01; d3=-2e-8; f=7000
function TemperatureHiComp (Value : word) : integer;
var
  I : word;
begin
  I := word(abs(integer (Value) - 7000));
  return (TemperatureHi (Value) - MulDivInt (I, I, 50000) div 100);
end;

//=== FLOATING POINT CONVERSION HELPER FUNCTIONS ===============================

{$IFDEF Float}  // only available if float defined (neat)

// These are "money no object" floating point equations straight from the datasheet.
// Low and high resolution versions are not included, both temperature and
// humidity conversion functions are linearity compensated.  If you want to use the
// low resolution operation or not use the non-linearity compensation then it is
// simple to fiddle with these functions after reading the data sheet.

// 14 bit temperature conversion and linearisation constants
const
  f  : float = 7000.0;
  d1 : float = -40.0;  // @ 5volts - see datasheets for effect of other voltages
  d2 : float = 0.01;
  d3 : float = -2.0E-8;
(*
// 12 bit temperature conversion and linearisation constants
const
  f  : float = 1750.0;
  d1 : float = -40.0;  // @ 5volts  - see datasheets for effect of other voltages
  d2 : float = 0.04;
  d3 : float = -3.2E-7;
*)

function Temperature (Value : word) : float;
begin
  return (d1 + d2 * float (Value) + d3 * sqr (float (Value) - f));
end;

// 12 bit humidity conversion and linearisation constants
const
  c1 : float = -4.0;
  c2 : float = 0.0405;
  c3 : float = -2.8e-6;
  t1 : float = 0.01;
  t2 : float = 0.00008;

(*
// 8 bit humidity conversion and linearisation constants
const
  c1 : float = -4.0;
  c2 : float = 0.648;
  c3 : float = -7.2e-4;
  t1 : float = 0.01;
  t2 : float = 0.00128;
*)

function Humidity (Value : word; Temp : float) : float;
begin
  return (c1 + c2 * float (Value) + c3 * sqr (float (Value)) + (Temp - 25) * (t1 + t2 * float (Value)));
end;

// Returns the dewpoint temperature in deg C (T must be in deg C)
// see the datasheet to understand how it works ...
function DewPoint (T, RH : float) : float;
var
  LogEW : float;
begin
  LogEW := 0.66077 + 7.5 * T /(237.3 + T) + log10(RH) - 2;
  return ((0.66077 - LogEW) * 237.3 / (LogEW - 8.16077));
end;

{$ENDIF}



